use crate::lib::screen::{Screen, BoardCommand};
use crate::lib::point::Point;
use crate::lib::entity::{Entities, Queen, Egg, Ant, Food};
use std::collections::HashSet;

#[derive(Clone)]
pub struct World {
    pub cleared: HashSet<Point>,
    pub occupied: HashSet<Point>,
    pub food: HashSet<u32>
}

impl World {
    pub fn new() -> World {
        World {
            cleared: HashSet::new(),
            occupied: HashSet::new(),
            food: HashSet::new()
        }
    }

    pub fn clear(&mut self, pos: Point) {
        self.cleared.insert(pos);
    }

    pub fn is_valid_movement(&self, pos: &Point, b: &Screen) -> bool {
        b.is_in_bounds(pos) && !self.occupied.contains(pos)
    }

    pub fn get_valid_movements(&self, pos: &Point, b: &Screen) -> Vec<Point> {
        let moves = b.get_valid_movements(pos);
        moves
            .iter()
            .filter(|p| !self.occupied.contains(p))
            .map(|p| p.clone())
            .collect()
    }

    pub fn update_occupied(&mut self, old: &Point, new: &Point) {
        self.occupied.remove(old);
        self.occupied.insert(*new);
    }
}

pub fn render(e: &Entities, w: &World, b: &Screen) {
    for c in w.cleared.iter() {
        b.render(c, "x");
    }

    for a in e.data.values() {
        a.render(b);
    }
}

pub fn simulate(e: &mut Entities, w: &mut World, b: &mut Screen) {
    let cmds: Vec<BoardCommand> = e
        .data
        .values_mut()
        .map(|a| { 
            a.plan(w);
            a.step(b, w) 
        })
        .collect();

    for cmd in cmds {
        match cmd {
            BoardCommand::Dig(pos) => {
                w.clear(pos);
            }
            BoardCommand::LayEgg(pos, id) => {
                let egg = Egg::new(pos.0, pos.1, id);
                e.add_entity(&egg);
            }
            BoardCommand::Hatch(egg_id, queen_id) => {
                let egg = e.data.remove(&egg_id);
                let pos = egg.unwrap().get_position();

                let q = e.data.get_mut(&queen_id).unwrap();
                let queen: &mut Queen = q.downcast_mut::<Queen>().unwrap();
                queen.egg_count -= 1;
                let ant = Ant::new(pos.0, pos.1);
                e.add_entity(&ant);
            }
            BoardCommand::SpawnFood(pos) => {
                let food = Food::new(pos.0, pos.1);
                e.add_entity(&food); 
            }
            BoardCommand::Noop => {}
        }
    }
}
